@@ -7,37 +7,32 @@ Analysis routines are not written in the ABF class itself. If useful, they
 are to be written in another file and imported as necessary.
 """
 
+from io import BufferedReader
+import pathlib
+from pyabf.abf2.dataSection import DataSection
 import pyabf.abfWriter
 import pyabf.stimulus
-import pyabf.abfHeaderDisplay
 
-from pyabf.abfHeader import BLOCKSIZE
-from pyabf.abfHeader import StringsIndexed
-from pyabf.abfHeader import StringsSection
-from pyabf.abfHeader import TagSection
-from pyabf.abfHeader import EpochSection
-from pyabf.abfHeader import EpochPerDACSection
-from pyabf.abfHeader import DACSection
-from pyabf.abfHeader import ADCSection
-from pyabf.abfHeader import ProtocolSection
-from pyabf.abfHeader import SynchArraySection
-from pyabf.abfHeader import SectionMap
-from pyabf.abfHeader import HeaderV2
-from pyabf.abfHeader import HeaderV1
+from pyabf.abf2.stringsSection import StringsSection
+from pyabf.abf2.tagSection import TagSection
+from pyabf.abf2.epochSection import EpochSection
+from pyabf.abf2.epochPerDacSection import EpochPerDACSection
+from pyabf.abf2.dacSection import DACSection
+from pyabf.abf2.adcSection import ADCSection
+from pyabf.abf2.protocolSection import ProtocolSection
+from pyabf.abf2.synchArraySection import SynchArraySection
+from pyabf.abf2.userListSection import UserListSection
+from pyabf.abf2.headerV2 import HeaderV2
+from pyabf.abf1.headerV1 import HeaderV1
+
+from pyabf.tools.abfHeaderDisplay import abfInfoPage
 
-import pyabf.abfHeader
 import os
-import sys
-import glob
 import time
-import datetime
 import numpy as np
 from pathlib import PureWindowsPath
 import hashlib
-
-import logging
-logging.basicConfig(level=logging.WARN)
-log = logging.getLogger(__name__)
+from typing import Union, List
 
 
 class ABF:
@@ -53,12 +48,43 @@ class ABF:
     abf.setSweep() then access abf.sweepX and abf.sweepY and similar values.
     """
 
-    def __init__(self, abfFilePath, loadData=True,
-                 cacheStimulusFiles=True, stimulusFileFolder=None):
+    def __init__(self,
+                 abfFilePath: Union[str, pathlib.Path],
+                 loadData: bool = True,
+                 cacheStimulusFiles: bool = True,
+                 stimulusFileFolder: bool = None):
+        """
+        Load header and sweep data from an ABF file.
+
+        ### Parameters
+
+        1. abfFilePath -- path to the ABF file
+
+        2. loadData -- whether or not to load sweep data values from the file immediately on instantiation.
+        Set this to False if you intent to iterate many ABF files rapidly and only inspect their headers.
+
+        3. cacheStimulusFiles -- Some ABF files use a source ABF as a stimulus file to control its DAC.
+        This module automatically loads the stimulus waveform from disk if it is available.
+        This parameter controls whether stimulus files should be stored in memory at the module level.
+        Set to False to prevent memory issues if you intend to iterate very large numbers of ABFs with
+        unique stimulus waveforms.
+
+        4. stimulusFileFolder -- Some ABF files use a source ABF as a stimulus file to control its DAC.
+        This module automatically loads the stimulus waveform from disk if it is available.
+        The ABF header contains an absolute file path to the stimulus file used to control the DAC.
+        If supplied, this path is used as an alternate search path to look for stimulus files with the same filename
+        in the case the original path does not exist on the machine loading the ABF.
+        """
+
+        if (isinstance(abfFilePath, pathlib.Path)):
+            abfFilePath = str(abfFilePath)
 
         if abfFilePath.lower().endswith(".atf"):
             raise Exception("use pyabf.ATF (not pyabf.ABF) for ATF files")
 
+        if (os.path.isdir(abfFilePath)):
+            raise Exception("path must be a path to a FILE not a FOLDER.")
+
         self._preLoadData = loadData
         self._cacheStimulusFiles = cacheStimulusFiles
 
@@ -73,21 +99,21 @@ class ABF:
         if not os.path.exists(self.abfFilePath):
             raise ValueError("ABF file does not exist: %s" % self.abfFilePath)
         self.abfID = os.path.splitext(os.path.basename(self.abfFilePath))[0]
-        log.debug(self.__repr__())
 
         with open(self.abfFilePath, 'rb') as fb:
 
-            # get a preliminary ABF version from the ABF file itself
+            # The first 4 bytes of the ABF indicates what type of file it is
             self.abfVersion = {}
-            self.abfVersion["major"] = pyabf.abfHeader.abfFileFormat(fb)
-            if not self.abfVersion["major"] in [1, 2]:
-                raise NotImplementedError("Invalid ABF file format")
-
-            # read the ABF header and bring its contents to the local namespace
-            if self.abfVersion["major"] == 1:
+            fb.seek(0)
+            fileSignature = fb.read(4).decode("ascii", errors='ignore')
+            if fileSignature == "ABF ":
+                self.abfVersion["major"] = 1
                 self._readHeadersV1(fb)
-            elif self.abfVersion["major"] == 2:
+            elif fileSignature == "ABF2":
+                self.abfVersion["major"] = 2
                 self._readHeadersV2(fb)
+            else:
+                raise NotImplementedError("Invalid ABF file format")
 
             # create more local variables based on the header data
             self._makeAdditionalVariables()
@@ -102,10 +128,12 @@ class ABF:
                 self.setSweep(0)
 
     def __str__(self):
-        """Return a string describing basic properties of the loaded ABF."""
+        """
+        Return a string describing basic properties of the loaded ABF.
+        """
 
         txt = """
-        ABF (version VERSN)
+        ABF (vVERSN)
         with CHNM channels (CHUNITS),
         sampled at RATEKHZ kHz,
         containing SWCNT sweeps,
@@ -117,7 +145,16 @@ class ABF:
             txt = txt.replace("  ", " ")
 
         # ABF version
-        txt = txt.replace("VERSN", self.abfVersionString)
+        # NOTE: self.versionString is improperly implemented but left to maintain back compatibility
+        realVersion = self.abfVersionString.replace(".", "")
+        realVersion = list(realVersion)
+        realVersion.insert(1, ".")
+        realVersion = "".join(realVersion)
+        if (realVersion[-1] == "0"):
+            realVersion = realVersion[:-1]
+        if (realVersion[-1] == "0"):
+            realVersion = realVersion[:-1]
+        txt = txt.replace("VERSN", realVersion)
 
         # channels
         txt = txt.replace("CHNM", str(self.channelCount))
@@ -149,10 +186,9 @@ class ABF:
         return txt
 
     def __repr__(self):
-        return 'ABFcore(abf="%s", loadData=%s)' % \
-            (self.abfFilePath, self._preLoadData)
+        return self.__str__() + f" path={self.abfFilePath}"
 
-    def _readHeadersV1(self, fb):
+    def _readHeadersV1(self, fb: BufferedReader):
         """Populate class variables from the ABF1 header."""
         assert self.abfVersion["major"] == 1
 
@@ -175,7 +211,17 @@ class ABF:
             self.abfFileComment = self._headerV1.sFileCommentNew
         else:
             self.abfFileComment = self._headerV1.sFileCommentOld
-        self.userList = None
+        self.nOperationMode = self._headerV1.nOperationMode
+        try:
+            self.userList = [float(x)
+                             for x in self._headerV1.sULParamValueList if x]
+        except:
+            self.userList = self._headerV1.sULParamValueList
+        self.userListEnable = self._headerV1.nULEnable
+        self.userListParamToVary = self._headerV1.nULParamToVary
+        self.userListParamToVaryName = [pyabf.names.getUserListParameterName(x)
+                                        for x in self.userListParamToVary]
+        self.userListRepeat = self._headerV1.nULRepeat
         _tagMult = self._headerV1.fADCSampleInterval / 1e6
         _tagMult = _tagMult / self._headerV1.nADCNumChannels
         self.tagComments = self._headerV1.sTagComment
@@ -184,7 +230,7 @@ class ABF:
 
         # data info
         self._nDataFormat = self._headerV1.nDataFormat
-        self.dataByteStart = self._headerV1.lDataSectionPtr*BLOCKSIZE
+        self.dataByteStart = self._headerV1.lDataSectionPtr*512
         self.dataByteStart += self._headerV1.nNumPointsIgnored
         self.dataPointCount = self._headerV1.lActualAcqLength
 
@@ -224,65 +270,80 @@ class ABF:
         self._dataOffset = [0]*self.channelCount
 
         for index, channel in enumerate(self.channelList):
-            self._dataGain[index] /= self._headerV1.fInstrumentScaleFactor[channel]
-            self._dataGain[index] /= self._headerV1.fSignalGain[channel]
-            self._dataGain[index] /= self._headerV1.fADCProgrammableGain[channel]
-            if self._headerV1.nTelegraphEnable[channel] == 1:
-                self._dataGain[index] /= self._headerV1.fTelegraphAdditGain[channel]
+            adcIndex = self._headerV1.nADCSamplingSeq[channel]
+            self._dataGain[index] /= self._headerV1.fInstrumentScaleFactor[adcIndex]
+            self._dataGain[index] /= self._headerV1.fSignalGain[adcIndex]
+            self._dataGain[index] /= self._headerV1.fADCProgrammableGain[adcIndex]
+            if self._headerV1.nTelegraphEnable[adcIndex] == 1:
+                self._dataGain[index] /= self._headerV1.fTelegraphAdditGain[adcIndex]
             self._dataGain[index] *= self._headerV1.fADCRange
             self._dataGain[index] /= self._headerV1.lADCResolution
-            self._dataOffset[index] += self._headerV1.fInstrumentOffset[channel]
-            self._dataOffset[index] -= self._headerV1.fSignalOffset[channel]
+            self._dataOffset[index] += self._headerV1.fInstrumentOffset[adcIndex]
+            self._dataOffset[index] -= self._headerV1.fSignalOffset[adcIndex]
 
-    def _readHeadersV2(self, fb):
+    def _readHeadersV2(self, fb: BufferedReader):
         """Populate class variables from the ABF2 header."""
 
         assert self.abfVersion["major"] == 2
 
         # read the headers out of the file
         self._headerV2 = HeaderV2(fb)
-        self._sectionMap = SectionMap(fb)
-        self._protocolSection = ProtocolSection(fb, self._sectionMap)
-        self._adcSection = ADCSection(fb, self._sectionMap)
-        self._dacSection = DACSection(fb, self._sectionMap)
-        self._epochPerDacSection = EpochPerDACSection(
-            fb, self._sectionMap)
-        self._epochSection = EpochSection(fb, self._sectionMap)
-        self._tagSection = TagSection(fb, self._sectionMap)
-        self._stringsSection = StringsSection(fb, self._sectionMap)
-        self._stringsIndexed = StringsIndexed(
-            self._headerV2, self._protocolSection, self._adcSection,
-            self._dacSection, self._stringsSection)
-        self._synchArraySection = SynchArraySection(fb, self._sectionMap)
+        self._protocolSection = ProtocolSection(fb)
+        self._dataSection = DataSection(fb)
+        self._adcSection = ADCSection(fb)
+        self._dacSection = DACSection(fb)
+        self._epochPerDacSection = EpochPerDACSection(fb)
+        self._epochSection = EpochSection(fb)
+        self._tagSection = TagSection(fb)
+        self._stringsSection = StringsSection(fb)
+        self._synchArraySection = SynchArraySection(fb)
+        self._userListSection = UserListSection(fb)
 
         # create useful variables at the class level
         self.abfVersion = self._headerV2.abfVersionDict
         self.abfVersionString = self._headerV2.abfVersionString
         self._fileGUID = self._headerV2.sFileGUID
-        self.creator = self._stringsIndexed.uCreatorName + \
-            " " + self._headerV2.creatorVersionString
+        self.creator = \
+            self._stringsSection._indexedStrings[self._headerV2.uCreatorNameIndex] + " " + \
+            self._headerV2.creatorVersionString
         self.creatorVersion = self._headerV2.creatorVersionDict
         self.creatorVersionString = self._headerV2.creatorVersionString
         self.abfDateTime = self._headerV2.abfDateTime
         self.abfDateTimeString = self._headerV2.abfDateTimeString
         self.holdingCommand = self._dacSection.fDACHoldingLevel
-        self.protocolPath = self._stringsIndexed.uProtocolPath
-        self.abfFileComment = self._stringsIndexed.lFileComment
+        self.protocolPath = self._stringsSection._indexedStrings[self._headerV2.uProtocolPathIndex]
+        self.abfFileComment = self._stringsSection._indexedStrings[
+            self._protocolSection.lFileCommentIndex]
+        self.nOperationMode = self._protocolSection.nOperationMode
+
+        # populate the user list
+        self.userList = None
+        self.userListEnable = self._userListSection.nULEnable
+        self.userListParamToVary = self._userListSection.nULParamToVary
+        self.userListParamToVaryName = [pyabf.names.getUserListParameterName(x)
+                                        for x in self.userListParamToVary]
+        self.userListRepeat = self._userListSection.nULRepeat
 
-        # attempt to populate the user list
-        firstBlockStrings = self._stringsSection.strings[0].split(b'\x00')
-        self.userList = firstBlockStrings[-2].decode("utf-8").split(",")
         try:
+            # This is the correct way, but it doesn't seem to work for every ABF.
+            # I think this is because there is a bug in the string indexer.
+            #self.userList = self._stringsIndexed.indexedStrings[self._userListSection.nStringIndex[0]]
+            #self.userList = [float(x) for x in self.userList.split(",")]
+
+            # This is weird but it's been in the code for a while and seems to work.
+            firstBlockStrings = self._stringsSection._stringsRaw[0].split(
+                b'\x00')
+            self.userList = firstBlockStrings[-2].decode("utf-8").split(",")
             self.userList = [float(x) for x in self.userList if x]
         except:
             self.userList = None
 
         # data info
         self._nDataFormat = self._headerV2.nDataFormat
-        self.dataByteStart = self._sectionMap.DataSection[0]*BLOCKSIZE
-        self.dataPointCount = self._sectionMap.DataSection[2]
-        self.dataPointByteSize = self._sectionMap.DataSection[1]
-        self.channelCount = self._sectionMap.ADCSection[2]
+        self.dataByteStart = self._dataSection._byteStart
+        self.dataPointCount = self._dataSection._entryCount
+        self.dataPointByteSize = self._dataSection._entrySize
+        self.channelCount = self._adcSection._entryCount
         self.dataRate = self._protocolSection.fADCSequenceInterval
         self.dataRate = int(1e6 / self.dataRate)
         self.dataSecPerPoint = 1.0 / self.dataRate
@@ -302,24 +363,31 @@ class ABF:
             self.tagTimesSec[i] = round(self.tagTimesSec[i], 5)
 
         # channel names
-        self.adcUnits = self._stringsIndexed.lADCUnits[:self.channelCount]
-        self.adcNames = self._stringsIndexed.lADCChannelName[:self.channelCount]
-        self.dacUnits = self._stringsIndexed.lDACChannelUnits[:self.channelCount]
-        self.dacNames = self._stringsIndexed.lDACChannelName[:self.channelCount]
+        self.adcNames = [self._stringsSection._indexedStrings[x]
+                         for x in self._adcSection.lADCChannelNameIndex[:self.channelCount]]
+        self.adcUnits = [self._stringsSection._indexedStrings[x]
+                         for x in self._adcSection.lADCUnitsIndex[:self.channelCount]]
+        self.dacNames = [self._stringsSection._indexedStrings[x]
+                         for x in self._dacSection.lDACChannelNameIndex[:self.channelCount]]
+        self.dacUnits = [self._stringsSection._indexedStrings[x]
+                         for x in self._dacSection.lDACChannelUnitsIndex[:self.channelCount]]
 
         # data scaling
         self._dataGain = [1]*self.channelCount
         self._dataOffset = [0]*self.channelCount
         for i in range(self.channelCount):
-            self._dataGain[i] /= self._adcSection.fInstrumentScaleFactor[i]
-            self._dataGain[i] /= self._adcSection.fSignalGain[i]
-            self._dataGain[i] /= self._adcSection.fADCProgrammableGain[i]
-            if self._adcSection.nTelegraphEnable[i] == 1:
-                self._dataGain[i] /= self._adcSection.fTelegraphAdditGain[i]
+            #adcIndex = self._adcSection.nADCSamplingSeq[i]
+            # NOTE: ADC sequence is handled inside the ADC section so it doesn't need to be handled here
+            adcIndex = i
+            self._dataGain[i] /= self._adcSection.fInstrumentScaleFactor[adcIndex]
+            self._dataGain[i] /= self._adcSection.fSignalGain[adcIndex]
+            self._dataGain[i] /= self._adcSection.fADCProgrammableGain[adcIndex]
+            if self._adcSection.nTelegraphEnable[adcIndex] == 1:
+                self._dataGain[i] /= self._adcSection.fTelegraphAdditGain[adcIndex]
             self._dataGain[i] *= self._protocolSection.fADCRange
             self._dataGain[i] /= self._protocolSection.lADCResolution
-            self._dataOffset[i] += self._adcSection.fInstrumentOffset[i]
-            self._dataOffset[i] -= self._adcSection.fSignalOffset[i]
+            self._dataOffset[i] += self._adcSection.fInstrumentOffset[adcIndex]
+            self._dataOffset[i] -= self._adcSection.fSignalOffset[adcIndex]
 
     def _makeAdditionalVariables(self):
         """create or touch-up version-nonspecific variables."""
@@ -395,13 +463,12 @@ class ABF:
         else:
             raise NotImplementedError("unknown data format")
 
-    def _loadAndScaleData(self, fb):
+    def _loadAndScaleData(self, fb: BufferedReader):
         """Load data from the ABF file and scale it by its scaleFactor."""
 
         # read the data from the ABF file
         fb.seek(self.dataByteStart)
-        raw = np.fromfile(fb, dtype=self._dtype,
-                          count=self.dataPointCount)
+        raw = np.fromfile(fb, dtype=self._dtype, count=self.dataPointCount)
         nRows = self.channelCount
         nCols = int(self.dataPointCount/self.channelCount)
         raw = np.reshape(raw, (nCols, nRows))
@@ -437,21 +504,21 @@ class ABF:
     @property
     def headerText(self):
         """Return all header information as a text-formatted string."""
-        return pyabf.abfHeaderDisplay.abfInfoPage(self).getText()
+        return abfInfoPage(self).getText()
 
     @property
     def headerMarkdown(self):
         """Return all header information as a markdown-formatted string."""
-        return pyabf.abfHeaderDisplay.abfInfoPage(self).generateMarkdown()
+        return abfInfoPage(self).generateMarkdown()
 
     @property
     def headerHTML(self):
         """Return all header information as a text-formatted string."""
-        return pyabf.abfHeaderDisplay.abfInfoPage(self).generateHTML()
+        return abfInfoPage(self).generateHTML()
 
     def headerLaunch(self):
         """Display ABF header information in the web browser."""
-        html = pyabf.abfHeaderDisplay.abfInfoPage(self).generateHTML()
+        html = abfInfoPage(self).generateHTML()
 
         # open a temp file, save HTML, launch it, then delete it
         import tempfile
@@ -461,51 +528,66 @@ class ABF:
 
         try:
             with open(tmpFilePath, 'w') as f:
-                log.info("creating a temporary webpage %s ..." % (tmpFilePath))
                 f.write(html)
-            log.info("launching file in a web browser ...")
             os.system(tmpFilePath)
         finally:
-            log.info("waiting a few seconds for the browser to launch...")
             time.sleep(3)  # give it time to display before deleting the file
             os.remove(tmpFilePath)
-            log.info("deleted %s" % (tmpFilePath))
 
-    def saveABF1(self, filename, sampleRateHz):
+    def saveABF1(self, filePath: Union[str, pathlib.Path]):
         """
-        Save this ABF file as an ABF1 file compatible with ClampFit and
-        MiniAnalysis. To create an ABF1 file from scratch (not starting from
-        an existing ABF file), see methods in the pyabf.abfWriter module.
+        Save this ABF file as an ABF1 file compatible with ClampFit and MiniAnalysis.
+        Not all header values are saved, but the minimum necessary are to read sweep data.
+
+        ### Parameters
+        1. filePath -- path of the ABF1 file to create
         """
-        filename = os.path.abspath(filename)
-        log.info("Saving ABF as ABF1 file: %s" % filename)
+        if (self.nOperationMode == 1):
+            raise Exception(
+                "saving ABFs with variable-length sweeps is not supported")
+
+        if isinstance(filePath, pathlib.Path):
+            filePath = str(filePath)
+
+        filePath = os.path.abspath(filePath)
         sweepData = np.empty((self.sweepCount, self.sweepPointCount))
         for sweep in self.sweepList:
             self.setSweep(sweep)
             sweepData[sweep] = self.sweepY
-        pyabf.abfWriter.writeABF1(sweepData, filename, sampleRateHz)
-        log.info("saved ABF1 file: %s" % filename)
+        pyabf.abfWriter.writeABF1(sweepData, filePath, self.dataRate)
 
     def launchInClampFit(self):
         """
         Launch the ABF in the default ABF viewing program (usually ClampFit) as
-        if it were double-clicked in the windows explorer. This will fail is
-        ClampFit is already open.
+        if it were double-clicked in the windows explorer. 
+
+        This assumes ClampFit is installed on your system and the default application
+        for opening ABF files.
+
+        This will fail is ClampFit is already open.
         """
         cmd = 'explorer.exe "%s"' % (self.abfFilePath)
         print("Launching %s.abf in ClampFit..." % (self.abfID))
         print(cmd)
         os.system(cmd)
 
-    def setSweep(self, sweepNumber, channel=0, absoluteTime=False,
-                 baseline=[None, None]):
+    def setSweep(self,
+                 sweepNumber: int,
+                 channel: int = 0,
+                 absoluteTime: bool = False,
+                 baseline: List[float] = [None, None]):
         """
         Args:
-            sweepNumber: sweep number to load (starting at 0)
-            channel: ABF channel (starting at 0)
-            absoluteTime: if False, sweepX always starts at 0.
-            baseline: a list of two times (seconds) the sweep will be baseline-
-                      subtraced to. Leave [None, None] to disable.
+
+        ### Parameters
+        1. sweepNumber -- sweep number to load (starting at 0)
+        2. channel -- ABF channel (starting at 0)
+        3. absoluteTime -- Whether sweepX should represent time in sweep or time in file.
+            If False, the first value of sweepX will always be 0.
+            If True, sweepX of successive sweeps will start at larger and larger values.
+        4. baseline -- If two times are provided (in seconds), the entire sweep will be subtracted
+            to the mean of the sweep between those two times (baseline subtraction).
+            Use [None, None] to disable automatic baseline subtraction.
         """
 
         # basic error checking
@@ -519,7 +601,6 @@ class ABF:
             raise ValueError(msg)
 
         if not "data" in (dir(self)):
-            log.debug("ABF data not preloaded. Loading now...")
             with open(self.abfFilePath, 'rb') as fb:
                 self._loadAndScaleData(fb)
 
@@ -578,13 +659,11 @@ class ABF:
 
         # default case is disabled
         if not hasattr(self, '_sweepBaselinePoints'):
-            log.debug("setSweep doesn't see baselinePoints, making False")
             self._sweepBaselinePoints = False
 
         # if baseline subtraction is used, apply it
         assert isinstance(baseline, list) and len(baseline) == 2
         if not None in baseline:
-            log.debug("setSweep is applying baseline subtraction")
             pt1, pt2 = [int(x*self.dataRate) for x in baseline]
             blVal = np.average(self.sweepY[pt1:pt2])
             self.sweepY = self.sweepY-blVal
@@ -659,7 +738,6 @@ class ABF:
 
     @property
     def fileGUID(self):
-        log.warning("fileGUID isn't truly unique (fileUUID is)")
         return self._fileGUID
 
     @property
